from __future__ import annotations

from typing import Any, Dict, List, Optional
import os
import time
import html as _html
from datetime import datetime

from ..utils import load_css, load_template, embed_favicon, embed_image, load_script
from ..accumulators import NumericAccumulator, CategoricalAccumulator, DatetimeAccumulator, BooleanAccumulator
from ..compute.infer import ColumnKinds
from .cards import render_numeric_card as _render_numeric_card, render_cat_card as _render_cat_card, render_dt_card as _render_dt_card, render_bool_card as _render_bool_card
from ..report_v2 import _resolve_pysuricata_version  # reuse version resolver


def _human_bytes(n: int) -> str:
    units = ["B", "KB", "MB", "GB", "TB", "PB"]
    size = float(max(0, n))
    for u in units:
        if size < 1024.0 or u == units[-1]:
            return f"{size:,.1f} {u}"
        size /= 1024.0


def render_html_snapshot(
    *,
    kinds: ColumnKinds,
    accs: Dict[str, Any],
    first_columns: List[str],
    row_kmv: Any,
    total_missing_cells: int,
    approx_mem_bytes: int,
    start_time: float,
    cfg: Any,
    report_title: Optional[str],
    sample_section_html: str,
) -> str:
    kinds_map = {
        **{name: ("numeric", accs[name]) for name in kinds.numeric},
        **{name: ("categorical", accs[name]) for name in kinds.categorical},
        **{name: ("datetime", accs[name]) for name in kinds.datetime},
        **{name: ("boolean", accs[name]) for name in kinds.boolean},
    }

    miss_list: List[tuple[str, float, int]] = []
    for name, (kind, acc) in kinds_map.items():
        miss = getattr(acc, "missing", 0)
        cnt = getattr(acc, "count", 0) + miss
        pct = (miss / cnt * 100.0) if cnt else 0.0
        miss_list.append((name, pct, miss))
    miss_list.sort(key=lambda t: t[1], reverse=True)
    top_missing_list = ""
    for col, pct, count in miss_list[:5]:
        severity_class = "low" if pct <= 5 else ("medium" if pct <= 20 else "high")
        top_missing_list += f"""
        <li class=\"missing-item\"> 
          <div class=\"missing-info\"> 
            <code class=\"missing-col\" title=\"{_html.escape(str(col))}\">{_html.escape(str(col))}</code>
            <span class=\"missing-stats\">{count:,} ({pct:.1f}%)</span>
          </div>
          <div class=\"missing-bar\"><div class=\"missing-fill {severity_class}\" style=\"width:{pct:.1f}%;\"></div></div>
        </li>
        """
    if not top_missing_list:
        top_missing_list = """
        <li class=\"missing-item\"><div class=\"missing-info\"><code class=\"missing-col\">None</code><span class=\"missing-stats\">0 (0.0%)</span></div><div class=\"missing-bar\"><div class=\"missing-fill low\" style=\"width:0%;\"></div></div></li>
        """

    n_rows = int(getattr(row_kmv, "rows", 0))
    n_cols = len(kinds_map)
    total_cells = n_rows * n_cols
    missing_overall = f"{total_missing_cells:,} ({(total_missing_cells/max(1,total_cells)*100):.1f}%)"
    dup_rows, dup_pct = row_kmv.approx_duplicates()
    duplicates_overall = f"{dup_rows:,} ({dup_pct:.1f}%)"

    constant_cols = 0
    high_card_cols = 0
    for name, (kind, acc) in kinds_map.items():
        if kind in ("numeric", "categorical"):
            u = acc._uniques.estimate() if hasattr(acc, "_uniques") else getattr(acc, "unique_est", 0)
        elif kind == "datetime":
            u = acc.unique_est
        else:
            present = (acc.true_n > 0) + (acc.false_n > 0)
            u = int(present)
        total = getattr(acc, "count", 0) + getattr(acc, "missing", 0)
        if u <= 1:
            constant_cols += 1
        if kind == "categorical" and n_rows:
            if (u / n_rows) > 0.5:
                high_card_cols += 1

    if kinds.datetime:
        mins, maxs = [], []
        for name in kinds.datetime:
            acc = accs[name]
            if acc._min_ts is not None: mins.append(acc._min_ts)
            if acc._max_ts is not None: maxs.append(acc._max_ts)
        if mins and maxs:
            date_min = datetime.utcfromtimestamp(min(mins) / 1_000_000_000).isoformat() + "Z"
            date_max = datetime.utcfromtimestamp(max(maxs) / 1_000_000_000).isoformat() + "Z"
        else:
            date_min = date_max = "—"
    else:
        date_min = date_max = "—"

    text_cols = len(kinds.categorical)
    avg_text_len_vals = [acc.avg_len for name, (k, acc) in kinds_map.items() if k == "categorical" and acc.avg_len is not None]
    avg_text_len = f"{(sum(avg_text_len_vals)/len(avg_text_len_vals)):.1f}" if avg_text_len_vals else "—"

    col_order = [c for c in list(first_columns) if c in kinds.numeric + kinds.categorical + kinds.datetime + kinds.boolean] or (kinds.numeric + kinds.categorical + kinds.datetime + kinds.boolean)
    all_cards_list: List[str] = []
    for name in col_order:
        acc = accs[name]
        if name in kinds.numeric:
            all_cards_list.append(_render_numeric_card(acc.finalize()))
        elif name in kinds.categorical:
            all_cards_list.append(_render_cat_card(acc.finalize()))
        elif name in kinds.datetime:
            all_cards_list.append(_render_dt_card(acc.finalize()))
        elif name in kinds.boolean:
            all_cards_list.append(_render_bool_card(acc.finalize()))
    variables_section_html = f"""
        <section id=\"vars\">  
          <span id=\"numeric-vars\" class=\"anchor-alias\"></span>
          <h2 class=\"section-title\">Variables</h2>
          <p class=\"muted small\">Analyzing {len(kinds.numeric)+len(kinds.categorical)+len(kinds.datetime)+len(kinds.boolean)} variables ({len(kinds.numeric)} numeric, {len(kinds.categorical)} categorical, {len(kinds.datetime)} datetime, {len(kinds.boolean)} boolean).</p>
          <div class=\"cards-grid\">{''.join(all_cards_list)}</div>
        </section>
    """
    sections_html = (sample_section_html or "") + variables_section_html

    module_dir = os.path.dirname(os.path.abspath(__file__))
    pkg_dir = os.path.dirname(module_dir)
    static_dir = os.path.join(pkg_dir, "static")
    template_dir = os.path.join(pkg_dir, "templates")
    template_path = os.path.join(template_dir, "report_template.html")
    template = load_template(template_path)
    css_path = os.path.join(static_dir, "css", "style.css")
    css_tag = load_css(css_path)
    script_path = os.path.join(static_dir, "js", "functionality.js")
    script_content = load_script(script_path)
    logo_light_path = os.path.join(static_dir, "images", "logo_suricata_transparent.png")
    logo_dark_path  = os.path.join(static_dir, "images", "logo_suricata_transparent_dark_mode.png")
    logo_light_img = embed_image(logo_light_path, element_id="logo-light", alt_text="Logo", mime_type="image/png")
    logo_dark_img  = embed_image(logo_dark_path,  element_id="logo-dark",  alt_text="Logo (dark)", mime_type="image/png")
    logo_html = f'<span id="logo">{logo_light_img}{logo_dark_img}</span>'
    favicon_path = os.path.join(static_dir, "images", "favicon.ico")
    favicon_tag = embed_favicon(favicon_path)

    end_time = time.time()
    duration_seconds = end_time - start_time
    report_date = datetime.now().strftime("%Y-%m-%d %H:%M:%S")
    pysuricata_version = _resolve_pysuricata_version()
    repo_url = "https://github.com/alvarodiez20/pysuricata"

    html = template.format(
      favicon=favicon_tag,
      css=css_tag,
      script=script_content,
      logo=logo_html,
      report_title= report_title or cfg.title,
      report_date=report_date,
      pysuricata_version=pysuricata_version,
      report_duration=f"{duration_seconds:.2f}",
      repo_url=repo_url,
      n_rows=f"{n_rows:,}",
      n_cols=f"{n_cols:,}",
      memory_usage=_human_bytes(approx_mem_bytes) if approx_mem_bytes else "—",
      missing_overall=missing_overall,
      duplicates_overall=duplicates_overall,
      numeric_cols=len(kinds.numeric),
      categorical_cols=len(kinds.categorical),
      datetime_cols=len(kinds.datetime),
      bool_cols=len(kinds.boolean),
      top_missing_list=top_missing_list,
      n_unique_cols=f"{n_cols:,}",
      constant_cols=f"{constant_cols:,}",
      high_card_cols=f"{high_card_cols:,}",
      date_min=date_min,
      date_max=date_max,
      text_cols=f"{text_cols:,}",
      avg_text_len=avg_text_len,
      dataset_sample_section=sections_html,
    )
    return html



"""Build the JSON-safe manifest from finalized accumulators."""

from __future__ import annotations

from typing import Any, Dict, List, Mapping, Sequence, Tuple

try:  # optional
    import pandas as pd  # type: ignore
except Exception:  # pragma: no cover
    pd = None  # type: ignore


def build_summary(
    kinds_map: Mapping[str, Tuple[str, Any]],
    col_order: Sequence[str],
    *,
    row_kmv: Any,
    total_missing_cells: int,
    n_rows: int,
    n_cols: int,
    miss_list: Sequence[Tuple[str, float, int]] = (),
) -> Mapping[str, Any]:
    """Construct a minimal, JSON-safe summary manifest.

    Parameters
    - kinds_map: name -> (kind, accumulator)
    - col_order: stable column order to iterate summaries
    - row_kmv: object exposing rows and approx_duplicates()
    - total_missing_cells: total missing across dataset
    - n_rows, n_cols: dataset shape estimates
    - miss_list: optional precomputed top-missing list [(name, pct, count)]
    """
    dataset_summary = {
        "rows_est": int(n_rows),
        "cols": int(n_cols),
        "missing_cells": int(total_missing_cells),
        "missing_cells_pct": (total_missing_cells / max(1, n_rows * n_cols) * 100.0) if (n_rows and n_cols) else 0.0,
        "duplicate_rows_est": int(row_kmv.approx_duplicates()[0]) if hasattr(row_kmv, "approx_duplicates") else 0,
        "duplicate_rows_pct_est": float(row_kmv.approx_duplicates()[1]) if hasattr(row_kmv, "approx_duplicates") else 0.0,
        "top_missing": [
            {"column": str(col), "pct": float(pct), "count": int(cnt)}
            for col, pct, cnt in (list(miss_list)[:5] if miss_list else [])
        ],
    }

    columns_summary: Dict[str, Dict[str, Any]] = {}
    for name in col_order:
        kind, acc = kinds_map[name]
        if kind == "numeric":
            s = acc.finalize()
            columns_summary[name] = {
                "type": "numeric",
                "count": s.count,
                "missing": s.missing,
                "unique_est": s.unique_est,
                "mean": s.mean,
                "std": s.std,
                "min": s.min,
                "q1": s.q1,
                "median": s.median,
                "q3": s.q3,
                "max": s.max,
                "zeros": s.zeros,
                "negatives": s.negatives,
                "outliers_iqr_est": s.outliers_iqr,
                "approx": bool(s.approx),
            }
        elif kind == "categorical":
            s = acc.finalize()
            columns_summary[name] = {
                "type": "categorical",
                "count": s.count,
                "missing": s.missing,
                "unique_est": s.unique_est,
                "top_items": s.top_items,
                "approx": bool(s.approx),
            }
        elif kind == "datetime":
            s = acc.finalize()
            columns_summary[name] = {
                "type": "datetime",
                "count": s.count,
                "missing": s.missing,
                "min_ts": s.min_ts,
                "max_ts": s.max_ts,
            }
        else:  # boolean
            s = acc.finalize()
            columns_summary[name] = {
                "type": "boolean",
                "count": s.count,
                "missing": s.missing,
                "true": s.true_n,
                "false": s.false_n,
            }

    return {"dataset": dataset_summary, "columns": columns_summary}




import pandas as pd

from pysuricata import profile, summarize, ReportConfig


def test_profile_returns_html_and_stats():
    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    rep = profile(df, config=ReportConfig())
    assert rep.html and isinstance(rep.html, str)
    assert rep.stats and isinstance(rep.stats, dict)
    assert rep.stats.get("dataset") is not None
    assert rep.stats.get("columns") is not None


def test_summarize_stats_only():
    df = pd.DataFrame({"a": [1, 2, 3], "b": [4, 5, 6]})
    stats = summarize(df)
    assert "dataset" in stats and "columns" in stats



from pysuricata.accumulators.categorical import CategoricalAccumulator


def test_categorical_accumulator_top_items_and_missing():
    acc = CategoricalAccumulator("c")
    data = ["a", "b", "a", None, "c", "a", "b", " ", ""]
    acc.update(data)
    s = acc.finalize()
    assert s.name == "c"
    assert s.count >= 6  # 3 non-missing a/b/a + c + a + b + " " + ""
    assert s.missing >= 1
    assert s.unique_est >= 1
    assert isinstance(s.top_items, list)
    # empty string counted
    assert s.empty_zero >= 1




from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Optional, Sequence, Tuple

import numpy as np

from .sketches import KMV, MisraGries, ReservoirSampler


@dataclass
class CategoricalSummary:
    name: str
    count: int
    missing: int
    unique_est: int
    top_items: List[Tuple[str, int]]
    approx: bool
    # extras for alignment
    mem_bytes: int = 0
    avg_len: Optional[float] = None
    len_p90: Optional[int] = None
    empty_zero: int = 0
    case_variants_est: int = 0
    trim_variants_est: int = 0
    dtype_str: str = "categorical"


class CategoricalAccumulator:
    def __init__(self, name: str, topk_k: int = 50, uniques_k: int = 2048) -> None:
        self.name = name
        self.count = 0
        self.missing = 0
        self._uniques = KMV(uniques_k)
        self._uniques_lower = KMV(uniques_k)
        self._uniques_strip = KMV(uniques_k)
        self.topk = MisraGries(topk_k)
        self._len_sum = 0
        self._len_n = 0
        self._len_sample = ReservoirSampler(5000)
        self._empty_zero = 0
        self._bytes_seen = 0
        self._dtype_str: str = "categorical"

    def update(self, arr: Sequence[Any]) -> None:
        for v in arr:
            if v is None:
                self.missing += 1
                continue
            if isinstance(v, float):
                try:
                    import math
                    if math.isnan(v):
                        self.missing += 1
                        continue
                except Exception:
                    pass
            self.count += 1
            s = str(v)
            self.topk.add(s, 1)
            self._uniques.add(s)
            try:
                self._uniques_lower.add(s.lower())
                self._uniques_strip.add(s.strip())
            except Exception:
                pass
            L = len(s)
            self._len_sum += L
            self._len_n += 1
            self._len_sample.add(float(L))
            if L == 0:
                self._empty_zero += 1
            self._bytes_seen += L  # approximate

    def finalize(self) -> CategoricalSummary:
        items = self.topk.items()[:10]
        # cast keys to str
        top_items = [(str(k), int(c)) for k, c in items]
        return CategoricalSummary(
            name=self.name,
            count=self.count,
            missing=self.missing,
            unique_est=self._uniques.estimate(),
            top_items=top_items,
            approx=not self._uniques.is_exact,
            mem_bytes=int(self._bytes_seen),
            avg_len=(self._len_sum / self._len_n) if self._len_n else None,
            len_p90=(int(np.quantile(self._len_sample.values(), 0.90)) if self._len_sample.values() else None),
            empty_zero=int(self._empty_zero),
            case_variants_est=max(0, int(self._uniques.estimate() - self._uniques_lower.estimate())),
            trim_variants_est=max(0, int(self._uniques.estimate() - self._uniques_strip.estimate())),
            dtype_str=self._dtype_str,
        )

    @property
    def avg_len(self) -> Optional[float]:
        return (self._len_sum / self._len_n) if self._len_n else None

    def set_dtype(self, dtype_str: str) -> None:
        try:
            self._dtype_str = str(dtype_str)
        except Exception:
            self._dtype_str = "categorical"




from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Optional, Sequence
import math


@dataclass
class BooleanSummary:
    name: str
    count: int
    missing: int
    true_n: int
    false_n: int
    mem_bytes: int = 0
    dtype_str: str = "boolean"


class BooleanAccumulator:
    def __init__(self, name: str) -> None:
        self.name = name
        self.count = 0
        self.missing = 0
        self.true_n = 0
        self.false_n = 0
        self._dtype_str: str = "boolean"
        self._mem_bytes: int = 0

    def update(self, arr: Sequence[Any]) -> None:
        for v in arr:
            if v is None or (isinstance(v, float) and math.isnan(v)):
                self.missing += 1
                continue
            if bool(v):
                self.true_n += 1
            else:
                self.false_n += 1
            self.count += 1

    def finalize(self) -> BooleanSummary:
        # Approximate memory: ~1 byte per boolean cell (very rough; marked as ≈ in UI)
        mem_bytes = int(self._mem_bytes if self._mem_bytes > 0 else (self.count + self.missing))
        return BooleanSummary(self.name, self.count, self.missing, self.true_n, self.false_n, mem_bytes, self._dtype_str)

    def set_dtype(self, dtype_str: str) -> None:
        try:
            self._dtype_str = str(dtype_str)
        except Exception:
            self._dtype_str = "boolean"

    def add_mem(self, n: int) -> None:
        try:
            self._mem_bytes += int(n)
        except Exception:
            pass




from datetime import datetime, timezone

from pysuricata.accumulators.datetime import DatetimeAccumulator


def _to_ns(dt: datetime) -> int:
    return int(dt.replace(tzinfo=timezone.utc).timestamp() * 1_000_000_000)


def test_datetime_accumulator_min_max_and_buckets():
    acc = DatetimeAccumulator("ts")
    dts = [
        _to_ns(datetime(2024, 1, 1, 0, 0, 0)),
        _to_ns(datetime(2024, 1, 1, 12, 0, 0)),
        _to_ns(datetime(2024, 1, 2, 13, 0, 0)),
        None,
    ]
    acc.update(dts)
    s = acc.finalize()
    assert s.name == "ts"
    assert s.count == 3
    assert s.missing == 1
    assert s.min_ts is not None and s.max_ts is not None
    assert s.min_ts <= s.max_ts
    assert len(s.by_hour) == 24
    assert len(s.by_dow) == 7
    assert len(s.by_month) == 12




"""pysuricata package exports.

Preferred high-level API:
    from pysuricata import profile, summarize, ReportConfig
"""

# High-level API wrappers
from .api import (
    Report,
    ReportConfig,
    ComputeOptions,
    RenderOptions,
    profile,
    summarize,
)

# Intentionally no public generate_report; use profile()/summarize() only.



from __future__ import annotations

from typing import Any, Iterable, Iterator, Optional, Sequence, Union


FrameLike = Any  # engine-native frame (e.g., pandas.DataFrame)


def iter_chunks(
    data: Union[FrameLike, Iterable[FrameLike]],
    *,
    chunk_size: Optional[int] = 200_000,
    columns: Optional[Sequence[str]] = None,
) -> Iterator[FrameLike]:
    """Yield pandas DataFrame chunks from in-memory objects only.

    Supports:
    - In-memory pandas.DataFrame (sliced by rows)
    - In-memory polars.DataFrame (native chunks, no pandas conversion)
    - An iterable of pandas DataFrames (pass-through)

    Notes:
    - This adapter yields pandas DataFrames as the common interchange format,
      which is what the compute pipeline consumes today.
    - Optional dependencies (pandas/polars) are imported lazily.
    """

    # Iterable of frames (assumed pandas DataFrames); pass-through
    try:
        if hasattr(data, "__iter__") and not hasattr(data, "__array__"):
            # Avoid mistaking numpy arrays for iterables of frames
            # Avoid strings/bytes which were handled above
            if not isinstance(data, (dict, list, tuple, set)):
                for ch in data:  # type: ignore[assignment]
                    yield ch
                return
    except Exception:
        pass

    # In-memory pandas DataFrame
    try:
        import pandas as pd  # type: ignore
        if isinstance(data, pd.DataFrame):
            n = len(data)
            if n == 0:
                return
            step = int(chunk_size or n)
            for i in range(0, n, step):
                df = data.iloc[i : i + step]
                if columns is not None:
                    # best-effort selection
                    try:
                        df = df[list(columns)]
                    except Exception:
                        pass
                yield df
            return
    except Exception:
        pass

    # In-memory polars DataFrame -> polars chunks
    try:
        import polars as pl  # type: ignore
        if isinstance(data, pl.DataFrame):
            step = int(chunk_size or len(data))
            n = data.height
            use_cols = list(columns) if columns is not None else None
            for i in range(0, n, step):
                ch = data.slice(i, min(step, n - i))
                if use_cols is not None:
                    # best-effort selection
                    try:
                        ch = ch.select(use_cols)
                    except Exception:
                        pass
                yield ch
            return
        if hasattr(data, "collect") and hasattr(data, "slice") and data.__class__.__name__ == "LazyFrame":  # lazy polars
            lf = data
            if columns is not None:
                try:
                    lf = lf.select(list(columns))
                except Exception:
                    pass
            df = lf.collect()
            yield from iter_chunks(df, chunk_size=chunk_size, columns=columns)
            return
    except Exception:
        pass

    raise TypeError(
        "Unsupported input for iter_chunks. Provide an in-memory pandas/polars DataFrame, or an iterable of pandas DataFrames."
    )



"""In-memory chunk adapters.

Exports a unified interface that yields pandas DataFrame chunks from
in-memory inputs (DataFrames or iterables of DataFrames).
"""

from .base import iter_chunks  # re-export convenience

__all__ = [
    "iter_chunks",
]



from __future__ import annotations

from typing import Any, Dict, List, Sequence, Tuple
import hashlib
import random


def _u64(x: bytes) -> int:
    """Return a 64-bit unsigned integer hash from bytes using SHA1.

    Fast enough and avoids external dependencies. Uses the first 8 bytes
    of the sha1 digest to build an unsigned 64-bit integer.
    """
    return int.from_bytes(hashlib.sha1(x).digest()[:8], "big", signed=False)


class KMV:
    """K-Minimum Values distinct counter (approximate uniques) without extra deps.

    Keep the k smallest 64-bit hashes of the observed values. If fewer than k items
    have been seen, |S| is exact uniques. Otherwise, estimate uniques as (k-1)/t,
    where t is the kth smallest hash normalized to (0,1].
    """

    __slots__ = ("k", "_values")

    def __init__(self, k: int = 2048) -> None:
        self.k = int(k)
        self._values: List[int] = []  # store as integers in [0, 2^64)

    def add(self, v: Any) -> None:
        if v is None:
            v = b"__NULL__"
        elif isinstance(v, bytes):
            pass
        else:
            v = str(v).encode("utf-8", "ignore")
        h = _u64(v)
        if len(self._values) < self.k:
            self._values.append(h)
            if len(self._values) == self.k:
                self._values.sort()
        else:
            # maintain k-smallest set (max-heap simulation via last element after sort)
            if h < self._values[-1]:
                # insert in sorted order (k is small)
                lo, hi = 0, self.k - 1
                while lo < hi:
                    mid = (lo + hi) // 2
                    if self._values[mid] < h:
                        lo = mid + 1
                    else:
                        hi = mid
                self._values.insert(lo, h)
                # trim to size
                del self._values[self.k]

    @property
    def is_exact(self) -> bool:
        return len(self._values) < self.k

    def estimate(self) -> int:
        n = len(self._values)
        if n == 0:
            return 0
        if n < self.k:
            # exact
            return n
        # normalize kth smallest to (0,1]
        kth = self._values[-1]
        t = (kth + 1) / 2**64
        if t <= 0:
            return n
        return max(n, int(round((self.k - 1) / t)))


class ReservoirSampler:
    """Reservoir sampler for numeric/datetime values to approximate quantiles/histograms."""

    __slots__ = ("k", "_buf", "_seen")

    def __init__(self, k: int = 20_000) -> None:
        self.k = int(k)
        self._buf: List[float] = []
        self._seen: int = 0

    def add_many(self, arr: Sequence[float]) -> None:
        for x in arr:
            self.add(float(x))

    def add(self, x: float) -> None:
        self._seen += 1
        if len(self._buf) < self.k:
            self._buf.append(x)
        else:
            j = random.randint(1, self._seen)
            if j <= self.k:
                self._buf[j - 1] = x

    def values(self) -> List[float]:
        return self._buf


class MisraGries:
    """Heavy hitters (top-K) with deterministic memory.

    Maintains up to k counters. Good for approximate top categories.
    """

    __slots__ = ("k", "counters")

    def __init__(self, k: int = 50) -> None:
        self.k = int(k)
        self.counters: Dict[Any, int] = {}

    def add(self, x: Any, w: int = 1) -> None:
        if x in self.counters:
            self.counters[x] += w
            return
        if len(self.counters) < self.k:
            self.counters[x] = w
            return
        # decrement all
        to_del = []
        for key in list(self.counters.keys()):
            self.counters[key] -= w
            if self.counters[key] <= 0:
                to_del.append(key)
        for key in to_del:
            del self.counters[key]

    def items(self) -> List[Tuple[Any, int]]:
        # items are approximate; a second pass could refine if needed
        return sorted(self.counters.items(), key=lambda kv: (-kv[1], str(kv[0])[:64]))




"""Chunk consumption and accumulator wiring for pandas chunks."""

from __future__ import annotations

from typing import Any, Dict, Iterable, List, Optional
import math
import warnings

import numpy as np

try:  # optional
    import pandas as pd  # type: ignore
except Exception:  # pragma: no cover
    pd = None  # type: ignore

from ..accumulators import (
    NumericAccumulator,
    BooleanAccumulator,
    DatetimeAccumulator,
    CategoricalAccumulator,
)
from .infer import ColumnKinds, infer_kind_for_series_pandas


def _to_numeric_array_pandas(s: "pd.Series") -> np.ndarray:  # type: ignore[name-defined]
    """Best-effort fast path to float64 NumPy array with NaN for invalid.

    - If the Series is already numeric (including pandas nullable ints),
      avoid the overhead of `pd.to_numeric` and go straight to NumPy.
    - Otherwise, coerce with `pd.to_numeric(errors='coerce')`.
    """
    try:
        # Fast path for numeric dtypes (exclude booleans)
        if pd is not None:
            from pandas.api import types as pdt  # type: ignore
            dt = getattr(s, "dtype", None)
            if dt is not None and not pdt.is_bool_dtype(dt) and pdt.is_numeric_dtype(dt):
                return s.to_numpy(dtype="float64", copy=False)
    except Exception:
        # Fall through to the coercion path on any failure
        pass
    try:
        ns = pd.to_numeric(s, errors="coerce")  # type: ignore[operator]
        return ns.to_numpy(dtype="float64", copy=False)
    except Exception:
        # Last resort: NumPy coercion (may be slower for object dtype)
        return np.asarray(getattr(s, "to_numpy", lambda: np.asarray(s))(), dtype="float64")


def _to_bool_array_pandas(s: "pd.Series") -> List[Optional[bool]]:  # type: ignore[name-defined]
    if str(s.dtype).startswith("bool"):
        arr = s.astype("boolean").tolist()
        return [None if x is pd.NA else bool(x) for x in arr]

    def _coerce(v: Any) -> Optional[bool]:
        if v is None or (isinstance(v, float) and math.isnan(v)):
            return None
        vs = str(v).strip().lower()
        if vs in {"true", "1", "t", "yes", "y"}:
            return True
        if vs in {"false", "0", "f", "no", "n"}:
            return False
        return None

    return [_coerce(v) for v in s.tolist()]


def _to_datetime_ns_array_pandas(s: "pd.Series") -> List[Optional[int]]:  # type: ignore[name-defined]
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        try:
            ds = pd.to_datetime(s, errors="coerce", utc=True, format="mixed")
        except TypeError:
            ds = pd.to_datetime(s, errors="coerce", utc=True)
    vals = ds.astype("int64", copy=False).tolist()
    NAT_INT = -9223372036854775808
    out: List[Optional[int]] = []
    for v in vals:
        out.append(None if v == NAT_INT else int(v))
    return out


def _to_categorical_iter_pandas(s: "pd.Series") -> Iterable[Any]:  # type: ignore[name-defined]
    return s.tolist()


def consume_chunk_pandas(df: "pd.DataFrame", accs: Dict[str, Any], kinds: ColumnKinds, logger: Optional["logging.Logger"] = None) -> None:  # type: ignore[name-defined]
    # 1) Create accumulators for columns not seen in the first chunk
    for name in df.columns:
        if name in accs:
            continue
        kind = infer_kind_for_series_pandas(df[name])
        if kind == "numeric":
            accs[name] = NumericAccumulator(name)
            kinds.numeric.append(name)
        elif kind == "boolean":
            accs[name] = BooleanAccumulator(name)
            kinds.boolean.append(name)
        elif kind == "datetime":
            accs[name] = DatetimeAccumulator(name)
            kinds.datetime.append(name)
        else:
            accs[name] = CategoricalAccumulator(name)
            kinds.categorical.append(name)
        if logger:
            logger.info("➕ discovered new column '%s' inferred as %s", name, kind)

    # 2) Feed accumulators for columns present in this chunk
    for name, acc in accs.items():
        if name not in df.columns:
            if logger:
                logger.debug("column '%s' not present in this chunk; skipping", name)
            continue
        s = df[name]
        if isinstance(acc, NumericAccumulator):
            arr = _to_numeric_array_pandas(s)
            acc.update(arr)
            # Track extremes with indices for this chunk
            try:
                finite = np.isfinite(arr)
                if finite.any():
                    vals = arr[finite]
                    idx = s.index.to_numpy()[finite]
                    if vals.size > 0:
                        k = min(5, vals.size)
                        part_min = np.argpartition(vals, k - 1)[:k]
                        pairs_min = [(idx[i], float(vals[i])) for i in part_min]
                        part_max = np.argpartition(-vals, k - 1)[:k]
                        pairs_max = [(idx[i], float(vals[i])) for i in part_max]
                        acc.update_extremes(pairs_min, pairs_max)
            except Exception:
                pass
        elif isinstance(acc, BooleanAccumulator):
            arr = _to_bool_array_pandas(s)
            acc.update(arr)
            try:
                acc.add_mem(int(s.memory_usage(deep=True)))
            except Exception:
                pass
        elif isinstance(acc, DatetimeAccumulator):
            arr = _to_datetime_ns_array_pandas(s)
            acc.update(arr)
            try:
                acc.add_mem(int(s.memory_usage(deep=True)))
            except Exception:
                pass
        elif isinstance(acc, CategoricalAccumulator):
            acc.update(_to_categorical_iter_pandas(s))



"""Kind inference utilities (engine-specific adapters can wrap here)."""

from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List
import re
import warnings

try:  # optional
    import pandas as pd  # type: ignore
except Exception:  # pragma: no cover
    pd = None  # type: ignore
try:  # optional
    import polars as pl  # type: ignore
except Exception:  # pragma: no cover
    pl = None  # type: ignore

@dataclass
class ColumnKinds:
    """Simple container for column names grouped by kind."""

    numeric: List[str] = field(default_factory=list)
    categorical: List[str] = field(default_factory=list)
    datetime: List[str] = field(default_factory=list)
    boolean: List[str] = field(default_factory=list)

    def __repr__(self) -> str:  # pragma: no cover - debug nicety
        return f"ColumnKinds(num={len(self.numeric)}, cat={len(self.categorical)}, dt={len(self.datetime)}, bool={len(self.boolean)})"


def infer_kinds_pandas(df: "pd.DataFrame") -> ColumnKinds:  # type: ignore[name-defined]
    kinds = ColumnKinds()
    for name, s in df.items():
        dt = str(getattr(s, "dtype", "object"))
        if re.search("int|float|^UInt|^Int|^Float", dt, re.I):
            kinds.numeric.append(name)
        elif re.search("bool", dt, re.I):
            kinds.boolean.append(name)
        elif re.search("datetime", dt, re.I):
            kinds.datetime.append(name)
        else:
            # try coercions on a small sample for robustness
            sample = s.head(10_000)
            if pd is not None:
                # datetime?
                with warnings.catch_warnings():
                    warnings.simplefilter("ignore", UserWarning)
                    try:
                        ds = pd.to_datetime(sample, errors="coerce", utc=True, format="mixed")
                    except TypeError:
                        ds = pd.to_datetime(sample, errors="coerce", utc=True)
                if ds.notna().sum() >= max(10, int(0.7 * len(sample))):
                    kinds.datetime.append(name)
                    continue
                ns = pd.to_numeric(sample, errors="coerce")
                if ns.notna().sum() >= max(10, int(0.7 * len(sample))):
                    kinds.numeric.append(name)
                    continue
            # boolean-like?
            uniq = set(map(str, sample.dropna().unique().tolist()))
            if 1 <= len(uniq) <= 2 and uniq.issubset({"True", "False", "0", "1", "true", "false"}):
                kinds.boolean.append(name)
            else:
                kinds.categorical.append(name)
    return kinds


def infer_kind_for_series_pandas(s: "pd.Series") -> str:  # type: ignore[name-defined]
    dt = str(getattr(s, "dtype", "object"))
    if re.search("int|float|^UInt|^Int|^Float", dt, re.I):
        return "numeric"
    if re.search("bool", dt, re.I):
        return "boolean"
    if re.search("datetime", dt, re.I):
        return "datetime"
    sample = s.head(10_000)
    if pd is not None:
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", UserWarning)
            try:
                ds = pd.to_datetime(sample, errors="coerce", utc=True, format="mixed")
            except TypeError:
                ds = pd.to_datetime(sample, errors="coerce", utc=True)
        if ds.notna().sum() >= max(10, int(0.7 * len(sample))):
            return "datetime"
        ns = pd.to_numeric(sample, errors="coerce")
        if ns.notna().sum() >= max(10, int(0.7 * len(sample))):
            return "numeric"
    uniq = set(map(str, sample.dropna().unique().tolist()))
    if 1 <= len(uniq) <= 2 and uniq.issubset({"True", "False", "0", "1", "true", "false"}):
        return "boolean"
    return "categorical"


def infer_kinds_polars(df: "pl.DataFrame") -> ColumnKinds:  # type: ignore[name-defined]
    kinds = ColumnKinds()
    if pl is None:
        return kinds
    for name in df.columns:
        dt = df.schema.get(name)
        try:
            if pl.datatypes.is_numeric(dt) and not pl.datatypes.is_boolean(dt):
                kinds.numeric.append(name)
            elif pl.datatypes.is_boolean(dt):
                kinds.boolean.append(name)
            elif isinstance(dt, pl.Datetime) or dt == pl.Datetime:
                kinds.datetime.append(name)
            else:
                kinds.categorical.append(name)
        except Exception:
            kinds.categorical.append(name)
    return kinds


def infer_kind_for_series_polars(s: "pl.Series") -> str:  # type: ignore[name-defined]
    if pl is None:
        return "categorical"
    dt = s.dtype
    try:
        if pl.datatypes.is_numeric(dt) and not pl.datatypes.is_boolean(dt):
            return "numeric"
        if pl.datatypes.is_boolean(dt):
            return "boolean"
        if isinstance(dt, pl.Datetime) or dt == pl.Datetime:
            return "datetime"
    except Exception:
        pass
    return "categorical"



from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, List, Optional, Sequence, Tuple
import math
from collections import Counter

import numpy as np

from .sketches import ReservoirSampler, KMV


@dataclass
class NumericSummary:
    name: str
    count: int
    missing: int
    unique_est: int
    mean: float
    std: float
    variance: float
    se: float
    cv: float
    gmean: float
    min: float
    q1: float
    median: float
    q3: float
    iqr: float
    mad: float
    skew: float
    kurtosis: float
    jb_chi2: float
    max: float
    zeros: int
    negatives: int
    outliers_iqr: int
    approx: bool
    inf: int
    # v2 extras (approximate)
    int_like: bool = False
    unique_ratio_approx: float = float("nan")
    hist_counts: Optional[List[int]] = None
    top_values: List[Tuple[float, int]] = field(default_factory=list)
    # carry a small reservoir sample to render histograms with axes (lin/log)
    sample_vals: Optional[List[float]] = None
    # quality/extras
    heap_pct: float = float("nan")
    gran_decimals: Optional[int] = None
    gran_step: Optional[float] = None
    bimodal: bool = False
    ci_lo: float = float("nan")
    ci_hi: float = float("nan")
    # alignment extras
    mem_bytes: int = 0
    mono_inc: bool = False
    mono_dec: bool = False
    dtype_str: str = "numeric"
    corr_top: List[Tuple[str, float]] = field(default_factory=list)
    sample_scale: float = 1.0
    # extremes with indices (global index labels where available)
    min_items: List[Tuple[Any, float]] = field(default_factory=list)
    max_items: List[Tuple[Any, float]] = field(default_factory=list)


class NumericAccumulator:
    """Streaming numeric stats + sampling for quantiles and histogram."""

    def __init__(self, name: str, sample_k: int = 20_000, uniques_k: int = 2048) -> None:
        self.name = name
        self.count = 0
        self.missing = 0
        self.zeros = 0
        self.negatives = 0
        self._min = math.inf
        self._max = -math.inf
        self._mean = 0.0
        self._m2 = 0.0  # for variance (Welford)
        self._m3 = 0.0
        self._m4 = 0.0
        self._log_sum_pos = 0.0
        self._pos_count = 0
        self._sample = ReservoirSampler(sample_k)
        self._uniques = KMV(uniques_k)
        self.inf = 0
        # track whether all observed values are integer-like (within tiny epsilon)
        self._int_like_all = True
        # streaming memory and monotonicity tracking
        self._bytes_seen = 0
        self._last_val: Optional[float] = None
        self._mono_inc = True
        self._mono_dec = True
        self._dtype_str: str = "numeric"
        self._corr_top: List[Tuple[str, float]] = []
        # Track extremes with indices across chunks (exact mins/maxs)
        self._min_pairs: List[Tuple[Any, float]] = []  # (idx, value)
        self._max_pairs: List[Tuple[Any, float]] = []

    def set_dtype(self, dtype_str: str) -> None:
        try:
            self._dtype_str = str(dtype_str)
        except Exception:
            self._dtype_str = "numeric"

    def set_corr_top(self, items: List[Tuple[str, float]]) -> None:
        self._corr_top = list(items or [])

    def update_extremes(self, pairs_min: List[Tuple[Any, float]], pairs_max: List[Tuple[Any, float]]) -> None:
        # Merge chunk minima
        for idx, v in pairs_min:
            self._min_pairs.append((idx, float(v)))
        # Keep only 5 smallest
        self._min_pairs.sort(key=lambda t: t[1])
        if len(self._min_pairs) > 5:
            del self._min_pairs[5:]
        # Merge chunk maxima
        for idx, v in pairs_max:
            self._max_pairs.append((idx, float(v)))
        # Keep only 5 largest
        self._max_pairs.sort(key=lambda t: -t[1])
        if len(self._max_pairs) > 5:
            del self._max_pairs[5:]

    def _combine_moments(self, n2: int, mean2: float, M2_2: float, M3_2: float, M4_2: float) -> None:
        """Merge another group's (n2, mean2, M2_2, M3_2, M4_2) into this accumulator."""
        if n2 <= 0:
            return
        n1 = self.count
        if n1 == 0:
            self.count = int(n2)
            self._mean = float(mean2)
            self._m2 = float(M2_2)
            self._m3 = float(M3_2)
            self._m4 = float(M4_2)
            return

        delta = mean2 - self._mean
        n = n1 + n2
        delta2 = delta * delta
        delta3 = delta2 * delta
        delta4 = delta2 * delta2

        M2 = self._m2 + M2_2 + delta2 * (n1 * n2 / n)
        M3 = (
            self._m3 + M3_2
            + delta3 * (n1 * n2 * (n1 - n2) / (n * n))
            + 3.0 * delta * (n1 * M2_2 - n2 * self._m2) / n
        )
        M4 = (
            self._m4 + M4_2
            + delta4 * (n1 * n2 * (n1 * n1 - n1 * n2 + n2 * n2) / (n ** 3))
            + 6.0 * delta2 * (n1 * n1 * M2_2 + n2 * n2 * self._m2) / (n * n)
            + 4.0 * delta * (n1 * M3_2 - n2 * self._m3) / n
        )

        self.count = int(n)
        self._mean = self._mean + (n2 / n) * delta
        self._m2 = float(M2)
        self._m3 = float(M3)
        self._m4 = float(M4)

    def update(self, arr: Sequence[Optional[float]]) -> None:
        """Update accumulator with a chunk of values.

        Preferred path: if `arr` is (or can be) a NumPy float array, compute
        per-chunk stats vectorially and merge them via `_combine_moments`.
        Fallback path: element-wise update (legacy behavior).
        """
        # --- Fast vectorized path (NumPy array of floats with NaN/±inf) ---
        try:
            a = arr if isinstance(arr, np.ndarray) else np.asarray(arr, dtype=float)
        except Exception:
            a = None  # triggers fallback

        if isinstance(a, np.ndarray) and np.issubdtype(a.dtype, np.floating):
            finite_mask = np.isfinite(a)  # False for NaN and ±inf
            if finite_mask.size == 0:
                return
            nan_mask = np.isnan(a)
            inf_mask = np.isinf(a)
            self.missing += int(nan_mask.sum())
            self.inf += int(inf_mask.sum())

            arr_f = a[finite_mask]
            if arr_f.size == 0:
                return

            # memory accounting (approximate: bytes of finite values processed)
            try:
                self._bytes_seen += int(arr_f.nbytes)
            except Exception:
                self._bytes_seen += int(arr_f.size * 8)

            # Basic counts and extremas on finite values
            n2 = int(arr_f.size)
            min2 = float(np.min(arr_f))
            max2 = float(np.max(arr_f))
            zeros2 = int((arr_f == 0.0).sum())
            neg2 = int((arr_f < 0.0).sum())

            # Geometric mean components (strictly positive)
            pos_mask = arr_f > 0.0
            pos_count2 = int(pos_mask.sum())
            log_sum_pos2 = float(np.log(arr_f[pos_mask]).sum()) if pos_count2 else 0.0

            # Integer-likeness tracking for this chunk
            int_like_chunk = bool(np.all(np.abs(arr_f - np.rint(arr_f)) < 1e-12))
            if self._int_like_all and not int_like_chunk:
                self._int_like_all = False

            # Monotonicity tracking against last value and within chunk (non-strict)
            try:
                if self._last_val is not None and arr_f.size > 0:
                    if arr_f[0] < self._last_val:
                        self._mono_inc = False
                    if arr_f[0] > self._last_val:
                        self._mono_dec = False
                if arr_f.size >= 2:
                    if np.any(arr_f[1:] < arr_f[:-1]):
                        self._mono_inc = False
                    if np.any(arr_f[1:] > arr_f[:-1]):
                        self._mono_dec = False
                if arr_f.size > 0:
                    self._last_val = float(arr_f[-1])
            except Exception:
                pass

            # Reservoir sample & uniques
            try:
                self._sample.add_many(arr_f.tolist())
            except Exception:
                for x in arr_f:
                    self._sample.add(float(x))
            for x in arr_f:
                self._uniques.add(float(x))

            # Per-chunk central moments (about mean)
            mean2 = float(arr_f.mean())
            d = arr_f - mean2
            M2_2 = float(np.sum(d * d))
            M3_2 = float(np.sum(d * d * d))
            M4_2 = float(np.sum(d * d * d * d))

            # Merge moments and update counters
            self._combine_moments(n2, mean2, M2_2, M3_2, M4_2)

            if min2 < self._min:
                self._min = min2
            if max2 > self._max:
                self._max = max2
            self.zeros += zeros2
            self.negatives += neg2
            self._log_sum_pos += log_sum_pos2
            self._pos_count += pos_count2
            return

        # --- Fallback: element-wise update (original logic) ---
        for v in arr:
            if v is None or (isinstance(v, float) and math.isnan(v)):
                self.missing += 1
                continue
            x = float(v)
            if math.isinf(x):
                self.inf += 1
                continue
            self._bytes_seen += 8
            if x == 0.0:
                self.zeros += 1
            if x < 0.0:
                self.negatives += 1
            if abs(x - round(x)) > 1e-12:
                self._int_like_all = False
            # monotonicity tracking
            if self._last_val is None:
                self._last_val = x
            else:
                if x < self._last_val:
                    self._mono_inc = False
                if x > self._last_val:
                    self._mono_dec = False
            self._last_val = x
            self._min = x if x < self._min else self._min
            self._max = x if x > self._max else self._max
            # Online update for mean, M2, M3, M4 (Pebay 2008)
            n1 = self.count
            n = n1 + 1
            delta = x - self._mean
            delta_n = delta / n
            delta_n2 = delta_n * delta_n
            term1 = delta * delta_n * n1
            m1_old = self._mean
            m2_old = self._m2
            m3_old = self._m3
            m4_old = self._m4
            m4 = m4_old + term1 * delta_n2 * (n * n - 3 * n + 3) + 6 * delta_n2 * m2_old - 4 * delta_n * m3_old
            m3 = m3_old + term1 * delta_n * (n - 2) - 3 * delta_n * m2_old
            m2 = m2_old + term1
            m1 = m1_old + delta_n
            self._mean = m1
            self._m2 = m2
            self._m3 = m3
            self._m4 = m4
            self.count = n
            if x > 0.0:
                self._log_sum_pos += math.log(x)
                self._pos_count += 1
            self._sample.add(x)
            self._uniques.add(x)

    def finalize(self) -> NumericSummary:
        if self.count > 1:
            var = self._m2 / (self.count - 1)
        else:
            var = float("nan")
        std = math.sqrt(var) if var >= 0 and not math.isnan(var) else float("nan")
        svals = sorted(self._sample.values())
        # v2 extras: unique ratio, histogram, top-5
        sample_n = len(svals)
        sample_unique = len(set(svals)) if svals else 0
        unique_ratio_approx = (sample_unique / sample_n) if sample_n else float("nan")

        def _q(p: float) -> float:
            if not svals:
                return float("nan")
            i = (len(svals) - 1) * p
            lo = math.floor(i)
            hi = math.ceil(i)
            if lo == hi:
                return float(svals[int(i)])
            return float(svals[lo] * (hi - i) + svals[hi] * (i - lo))

        q1 = _q(0.25)
        med = _q(0.5)
        q3 = _q(0.75)
        iqr = (q3 - q1) if all(map(math.isfinite, [q1, q3])) else float("nan")
        # MAD from sample (|x - median| median)
        if svals and math.isfinite(med):
            mad = float(np.median([abs(v - med) for v in svals]))
        else:
            mad = float("nan")
        # Histogram preview from sample (25 bins, linear scale)
        if svals:
            try:
                counts, _edges = np.histogram(svals, bins=25)
                hist_counts = [int(c) for c in counts.tolist()]
            except Exception:
                hist_counts = None
        else:
            hist_counts = None
        # Outliers via IQR on sample (scaled by sampling rate)
        if math.isfinite(iqr):
            lo, hi = q1 - 1.5 * iqr, q3 + 1.5 * iqr
        else:
            lo, hi = float("nan"), float("nan")
        outliers_iqr = 0
        if all(map(math.isfinite, [lo, hi])):
            if sample_n:
                out_s = sum(1 for x in svals if x < lo or x > hi)
                outliers_iqr = int(round(out_s * (self.count / sample_n)))
        unique_est = self._uniques.estimate()
        approx = not self._uniques.is_exact
        # Standard error & coefficient of variation
        se = (std / math.sqrt(self.count)) if (self.count > 1 and not math.isnan(std)) else float("nan")
        cv = (std / self._mean) if (self.count and not math.isnan(std) and self._mean != 0) else float("nan")
        # Geometric mean (only valid for strictly positive data)
        if self._min > 0 and self.count > 0 and self._pos_count == self.count:
            gmean = math.exp(self._log_sum_pos / self.count)
        else:
            gmean = float("nan")
        # Skewness, kurtosis (excess), and Jarque–Bera
        if self.count > 2 and self._m2 > 0 and math.isfinite(self._m2):
            skew = (math.sqrt(self.count) * self._m3) / (self._m2 ** 1.5)
        else:
            skew = float("nan")
        if self.count > 3 and self._m2 > 0 and math.isfinite(self._m2):
            kurt = (self.count * self._m4) / (self._m2 * self._m2) - 3.0
        else:
            kurt = float("nan")
        if all(map(lambda v: isinstance(v, (int, float)) and not math.isnan(v), [self.count, skew, kurt])) and self.count > 3:
            jb = float(self.count / 6.0 * (skew * skew + 0.25 * (kurt * kurt)))
        else:
            jb = float("nan")
        # Approximate Top-5 for discrete integer-like series
        top_values: List[Tuple[float, int]] = []
        try:
            # Heuristic: treat as discrete if all seen are int-like and small-ish cardinality
            is_discrete = bool(self._int_like_all and ((unique_est <= max(1, min(50, int(0.05 * max(1, self.count))))) or (sample_unique and (sample_unique <= 50 and unique_ratio_approx <= 0.05))))
            if is_discrete and svals:
                scale = (self.count / sample_n) if sample_n else 1.0
                ctr = Counter(svals)
                top = ctr.most_common(5)
                top_values = [(float(round(v)), int(round(c * scale))) for v, c in top]
        except Exception:
            top_values = []

        # 95% CI for mean
        if isinstance(se, float) and math.isfinite(se):
            ci_lo = float(self._mean - 1.96 * se)
            ci_hi = float(self._mean + 1.96 * se)
        else:
            ci_lo = float("nan"); ci_hi = float("nan")

        # Heap detection heuristic (sample-based)
        heap_pct = float("nan")
        try:
            if svals and len(svals) >= 25:
                # Look for mass at .0 or .5 boundaries
                if self._int_like_all:
                    # proportion at exact integers
                    ints = sum(1 for v in svals if abs(v - round(v)) <= 1e-12)
                    heap_pct = ints / len(svals) * 100.0
                else:
                    frac = [abs(v - round(v)) for v in svals]
                    # also catch .5 rounding
                    half = [abs((v - math.floor(v)) - 0.5) for v in svals]
                    tol = 0.002
                    hits = sum(1 for a,b in zip(frac, half) if (a <= tol) or (b <= tol))
                    heap_pct = hits / len(svals) * 100.0
        except Exception:
            pass

        # Granularity (decimals) and step size (approx)
        gran_decimals = None
        gran_step = None
        try:
            if svals:
                if not self._int_like_all:
                    # estimate modal count of decimals from string repr
                    decs = []
                    for x in svals[: min(50_000, len(svals))]:
                        xs = str(x)
                        if "." in xs:
                            decs.append(len(xs.split(".")[-1]))
                        else:
                            decs.append(0)
                    if decs:
                        # mode
                        counts = Counter(decs)
                        gran_decimals = int(max(counts.items(), key=lambda kv: (kv[1], -kv[0]))[0])
                # step ≈ median of positive diffs between unique sorted values (rounded if we have decimals)
                arr_g = np.unique(np.asarray(svals, dtype=float))
                if gran_decimals is not None and gran_decimals >= 0:
                    arr_g = np.round(arr_g, int(min(12, gran_decimals)))
                diffs = np.diff(np.sort(arr_g))
                diffs = diffs[np.isfinite(diffs) & (diffs > 0)]
                if diffs.size > 0:
                    gran_step = float(np.median(diffs))
        except Exception:
            pass

        # Bimodality hint via valley detection on binned counts (sample-based)
        bimodal = False
        try:
            if svals and len(svals) >= 10:
                nb = min(25, max(10, int(math.sqrt(len(svals)))))
                c_bi, _e_bi = np.histogram(svals, bins=nb)
                if c_bi.size >= 3:
                    valley = (c_bi[1:-1] < c_bi[:-2]) & (c_bi[1:-1] < c_bi[2:])
                    bimodal = bool((valley & (c_bi[1:-1] <= 0.8 * c_bi.max())).sum() >= 1)
        except Exception:
            bimodal = False

        sample_scale = (self.count / sample_n) if sample_n else 1.0
        return NumericSummary(
            name=self.name,
            count=self.count,
            missing=self.missing,
            unique_est=unique_est,
            mean=self._mean if self.count else float("nan"),
            std=std,
            variance=var,
            se=se,
            cv=cv,
            gmean=gmean,
            min=self._min if self.count else float("nan"),
            q1=q1,
            median=med,
            q3=q3,
            iqr=iqr,
            mad=mad,
            skew=skew,
            kurtosis=kurt,
            jb_chi2=jb,
            max=self._max if self.count else float("nan"),
            zeros=self.zeros,
            negatives=self.negatives,
            outliers_iqr=outliers_iqr,
            approx=approx,
            inf=self.inf,
            int_like=bool(self._int_like_all),
            unique_ratio_approx=unique_ratio_approx,
            hist_counts=hist_counts,
            top_values=top_values,
            sample_vals=svals,
            heap_pct=heap_pct,
            gran_decimals=gran_decimals,
            gran_step=gran_step,
            bimodal=bimodal,
            ci_lo=ci_lo,
            ci_hi=ci_hi,
            mem_bytes=int(self._bytes_seen),
            mono_inc=bool(self._mono_inc),
            mono_dec=bool(self._mono_dec),
            dtype_str=self._dtype_str,
            corr_top=list(self._corr_top),
            sample_scale=float(sample_scale),
            min_items=list(self._min_pairs),
            max_items=list(self._max_pairs),
        )




import pytest

try:
    import polars as pl  # type: ignore
except Exception:  # pragma: no cover
    pl = None  # type: ignore

from pysuricata.api import profile, ReportConfig


@pytest.mark.skipif(pl is None, reason="polars not installed")
def test_profile_polars_basic():
    df = pl.DataFrame({
        "a": [1, 2, 3, 4, 5],
        "b": [10.0, 20.0, None, 40.0, 50.0],
        "c": [True, False, True, None, False],
        "d": pl.date_range(low=pl.datetime(2024,1,1), high=pl.datetime(2024,1,5), interval="1d", eager=True),
        "e": ["x", "y", "x", "z", None],
    })
    rep = profile(df, config=ReportConfig())
    assert rep.html and isinstance(rep.html, str)
    assert rep.stats and isinstance(rep.stats, dict)
    assert rep.stats.get("dataset") is not None



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, List, Optional, Sequence
from datetime import datetime

from .sketches import KMV, ReservoirSampler


@dataclass
class DatetimeSummary:
    name: str
    count: int
    missing: int
    min_ts: Optional[int]
    max_ts: Optional[int]
    by_hour: List[int]  # 24 counts
    by_dow: List[int]   # 7 counts, Monday=0
    by_month: List[int] # 12 counts, Jan=1 index => store 12-length
    # v2 additions
    dtype_str: str = "datetime"
    mono_inc: bool = False
    mono_dec: bool = False
    mem_bytes: int = 0
    sample_ts: Optional[List[int]] = None
    sample_scale: float = 1.0


class DatetimeAccumulator:
    def __init__(self, name: str) -> None:
        self.name = name
        self.count = 0
        self.missing = 0
        self._min_ts: Optional[int] = None
        self._max_ts: Optional[int] = None
        self.by_hour = [0] * 24
        self.by_dow = [0] * 7
        self.by_month = [0] * 12
        self._uniques = KMV(2048)
        self._dtype_str: str = "datetime"
        self._mem_bytes: int = 0
        self._sample = ReservoirSampler(20_000)
        self._last_ts: Optional[int] = None
        self._mono_inc = True
        self._mono_dec = True

    def update(self, arr_ns: Sequence[Optional[int]]) -> None:
        # arr_ns: integer nanoseconds since epoch; None/NaN for missing
        for v in arr_ns:
            if v is None or (isinstance(v, float) and (v != v)):
                self.missing += 1
                continue
            ts = int(v)
            self.count += 1
            if self._min_ts is None or ts < self._min_ts:
                self._min_ts = ts
            if self._max_ts is None or ts > self._max_ts:
                self._max_ts = ts
            self._uniques.add(ts)
            # sample and monotonicity
            self._sample.add(float(ts))
            if self._last_ts is not None:
                if ts < self._last_ts:
                    self._mono_inc = False
                if ts > self._last_ts:
                    self._mono_dec = False
            self._last_ts = ts
            # breakdowns (UTC)
            try:
                dt = datetime.utcfromtimestamp(ts / 1_000_000_000)
                self.by_hour[dt.hour] += 1
                self.by_dow[dt.weekday()] += 1
                self.by_month[dt.month - 1] += 1
            except Exception:
                # ignore ill-formed
                pass

    def finalize(self) -> DatetimeSummary:
        svals = [int(x) for x in self._sample.values()]
        sample_scale = (self.count / len(svals)) if svals else 1.0
        return DatetimeSummary(
            name=self.name,
            count=self.count,
            missing=self.missing,
            min_ts=self._min_ts,
            max_ts=self._max_ts,
            by_hour=self.by_hour,
            by_dow=self.by_dow,
            by_month=self.by_month,
            dtype_str=self._dtype_str,
            mono_inc=bool(self._mono_inc),
            mono_dec=bool(self._mono_dec),
            mem_bytes=int(self._mem_bytes),
            sample_ts=svals,
            sample_scale=float(sample_scale),
        )

    @property
    def unique_est(self) -> int:
        return self._uniques.estimate()

    def set_dtype(self, dtype_str: str) -> None:
        try:
            self._dtype_str = str(dtype_str)
        except Exception:
            self._dtype_str = "datetime"

    def add_mem(self, n: int) -> None:
        try:
            self._mem_bytes += int(n)
        except Exception:
            pass




from __future__ import annotations

from typing import Any, List, Optional, Tuple, Sequence
import html as _html
import math

import numpy as np

try:  # optional
    import pandas as pd  # type: ignore
except Exception:  # pragma: no cover
    pd = None  # type: ignore


def _safe_col_id(name: str) -> str:
    return "col_" + "".join(ch if str(ch).isalnum() else "_" for ch in str(name))


def _human_bytes(n: int) -> str:
    units = ["B", "KB", "MB", "GB", "TB", "PB"]
    size = float(max(0, n))
    for u in units:
        if size < 1024.0 or u == units[-1]:
            return f"{size:,.1f} {u}"
        size /= 1024.0


def _fmt_num(x: Optional[float]) -> str:
    if x is None:
        return "—"
    try:
        if isinstance(x, float) and (x != x):  # NaN
            return "NaN"
        return f"{x:,.4g}"
    except Exception:
        return str(x)


def _svg_empty(css_class: str, width: int, height: int, aria_label: str = "no data") -> str:
    return f'<svg class="{css_class}" width="{width}" height="{height}" viewBox="0 0 {width} {height}" role="img" aria-label="{aria_label}"></svg>'


def _nice_ticks(vmin: float, vmax: float, n: int = 5):
    if not np.isfinite(vmin) or not np.isfinite(vmax):
        return [0, 1], 1
    if vmin == vmax:
        vmax = vmin + 1
    rng = vmax - vmin
    raw_step = rng / max(1, n)
    def nice_num(rng: float, do_round: bool = True) -> float:
        if rng <= 0 or not np.isfinite(rng):
            return 1.0
        exp = math.floor(math.log10(rng))
        frac = rng / (10 ** exp)
        if do_round:
            if frac < 1.5:
                nice = 1
            elif frac < 3:
                nice = 2
            elif frac < 7:
                nice = 5
            else:
                nice = 10
        else:
            if frac <= 1:
                nice = 1
            elif frac <= 2:
                nice = 2
            elif frac <= 5:
                nice = 5
            else:
                nice = 10
        return nice * (10 ** exp)
    step = nice_num(raw_step, True)
    start = math.floor(vmin / step) * step
    ticks = []
    x = start
    while x <= vmax + 1e-9:
        ticks.append(float(x))
        x += step
    return ticks, step


def _fmt_tick(v: float, step: float) -> str:
    if abs(v) >= 1000:
        return f"{v:,.0f}"
    if step >= 1:
        return f"{v:.0f}"
    if step >= 0.1:
        return f"{v:.1f}"
    if step >= 0.01:
        return f"{v:.2f}"
    try:
        return f"{v:.4g}"
    except Exception:
        return str(v)


def _fmt_compact(x) -> str:
    try:
        if x is None:
            return "—"
        if isinstance(x, float) and (math.isnan(x) or math.isinf(x)):
            return "—"
    except Exception:
        pass
    try:
        return f"{x:.4g}"
    except Exception:
        try:
            return f"{float(x):.4g}"
        except Exception:
            return str(x)


def _build_hist_svg_from_vals(
    base_title: str,
    vals: Sequence[float],
    *,
    bins: int = 25,
    width: int = 420,
    height: int = 160,
    margin_left: int = 45,
    margin_bottom: int = 36,
    margin_top: int = 8,
    margin_right: int = 8,
    scale: str = "lin",
    scale_count: float = 1.0,
    x_min_override: Optional[float] = None,
    x_max_override: Optional[float] = None,
) -> str:
    arr = np.asarray(vals, dtype=float)
    arr = arr[np.isfinite(arr)]
    if arr.size == 0:
        return _svg_empty("hist-svg", width, height)
    if scale == "log":
        arr = arr[arr > 0]
        if arr.size == 0:
            return _svg_empty("hist-svg", width, height)
        arr = np.log10(arr)

    x_min, x_max = float(np.min(arr)), float(np.max(arr))
    if x_min_override is not None and np.isfinite(x_min_override):
        x_min = float(x_min_override)
    if x_max_override is not None and np.isfinite(x_max_override):
        x_max = float(x_max_override)
    if x_min == x_max:
        x_min -= 0.5
        x_max += 0.5

    counts, edges = np.histogram(arr, bins=int(bins), range=(x_min, x_max))
    counts_scaled = np.maximum(0, np.round(counts * max(1.0, float(scale_count)))).astype(int)
    y_max = int(max(1, counts_scaled.max()))
    total_n = int(counts_scaled.sum()) if counts_scaled.size else 0

    iw = width - margin_left - margin_right
    ih = height - margin_top - margin_bottom

    def sx(x):
        return margin_left + (x - x_min) / (x_max - x_min) * iw

    def sy(y):
        return margin_top + (1 - y / y_max) * ih

    x_ticks, x_step = _nice_ticks(x_min, x_max, 6)
    xt = [x for x in x_ticks if x >= x_min - 1e-9 and x <= x_max + 1e-9]
    if not xt or abs(xt[0] - x_min) > 1e-9:
        xt = [x_min] + [x for x in xt if x > x_min]
    x_ticks = xt
    y_ticks, y_step = _nice_ticks(0, y_max, 5)

    parts = [
        f'<svg class="hist-svg" width="{width}" height="{height}" viewBox="0 0 {width} {height}" role="img" aria-label="Histogram">',
        '<g class="plot-area">'
    ]

    for i, c in enumerate(counts_scaled):
        x0 = edges[i]
        x1 = edges[i + 1]
        x = sx(x0)
        w = max(1.0, sx(x1) - sx(x0) - 1.0)
        y = sy(int(c))
        h = (margin_top + ih) - y
        pct = (c / total_n * 100.0) if total_n else 0.0
        parts.append(
            f'<rect class="bar" x="{x:.2f}" y="{y:.2f}" width="{w:.2f}" height="{h:.2f}" rx="1" ry="1" '
            f'data-count="{int(c)}" data-pct="{pct:.1f}" data-x0="{_fmt_compact(x0)}" data-x1="{_fmt_compact(x1)}">'
            f'<title>{int(c)} rows ({pct:.1f}%)&#10;[{_fmt_compact(x0)} – {_fmt_compact(x1)}]</title>'
            f'</rect>'
        )
    parts.append('</g>')

    x_axis_y = margin_top + ih
    parts.append(f'<line class="axis" x1="{margin_left}" y1="{x_axis_y}" x2="{margin_left + iw}" y2="{x_axis_y}"></line>')
    parts.append(f'<line class="axis" x1="{margin_left}" y1="{margin_top}" x2="{margin_left}" y2="{x_axis_y}"></line>')

    for xt in x_ticks:
        px = sx(xt)
        parts.append(f'<line class="tick" x1="{px}" y1="{x_axis_y}" x2="{px}" y2="{x_axis_y + 4}"></line>')
        parts.append(f'<text class="tick-label" x="{px}" y="{x_axis_y + 14}" text-anchor="middle">{_fmt_tick(xt, x_step)}</text>')
    for yt in y_ticks:
        py = sy(yt)
        parts.append(f'<line class="tick" x1="{margin_left - 4}" y1="{py}" x2="{margin_left}" y2="{py}"></line>')
        parts.append(f'<text class="tick-label" x="{margin_left - 6}" y="{py + 3}" text-anchor="end">{_fmt_tick(yt, y_step)}</text>')

    x_title = f"log10({base_title})" if scale == "log" else base_title
    parts.append(f'<text class="axis-title x" x="{margin_left + iw/2:.2f}" y="{x_axis_y + 28}" text-anchor="middle">{x_title}</text>')
    parts.append(f'<text class="axis-title y" transform="translate({margin_left - 36},{margin_top + ih/2:.2f}) rotate(-90)" text-anchor="middle">Count</text>')

    parts.append('</svg>')
    return ''.join(parts)


def render_numeric_card(s: Any) -> str:
    # For now, import and delegate to the implementation in report_v2 to keep parity
    # This shim can be expanded to fully isolate rendering logic if needed.
    # Avoid circular import at module import time
    from ..report_v2 import _render_numeric_card as _impl  # type: ignore
    return _impl(s)


def render_dt_card(s: Any) -> str:
    from ..report_v2 import _render_dt_card as _impl  # type: ignore
    return _impl(s)


def _build_cat_bar_svg_from_items(
    items: List[Tuple[str, int]],
    total: int,
    *,
    width: int = 420,
    height: int = 160,
    margin_top: int = 8,
    margin_bottom: int = 8,
    margin_left_min: int = 120,
    margin_right: int = 12,
    scale: str = "count",
) -> str:
    if total <= 0 or not items:
        return _svg_empty("cat-svg", width, height)
    labels = [_html.escape(str(k)) for k, _ in items]
    counts = [int(c) for _, c in items]
    pcts = [(c / total * 100.0) for c in counts]

    max_label_len = max((len(l) for l in labels), default=0)
    char_w = 7
    gutter = max(60, min(180, char_w * min(max_label_len, 28) + 16))
    mleft = max(margin_left_min, gutter)

    n = len(labels)
    iw = width - mleft - margin_right
    ih = height - margin_top - margin_bottom
    if n <= 0 or iw <= 0 or ih <= 0:
        return _svg_empty("cat-svg", width, height)
    bar_gap = 6
    bar_h = max(4, (ih - bar_gap * (n - 1)) / max(n, 1))

    if scale == "pct":
        vmax = max(pcts) or 1.0
        values = pcts
    else:
        vmax = float(max(counts)) or 1.0
        values = counts

    def sx(v: float) -> float:
        return mleft + (v / vmax) * iw

    parts = [f'<svg class="cat-svg" width="{width}" height="{height}" viewBox="0 0 {width} {height}" role="img" aria-label="Top categories">']
    for i, (label, c, p, val) in enumerate(zip(labels, counts, pcts, values)):
        y = margin_top + i * (bar_h + bar_gap)
        x0 = mleft
        x1 = sx(float(val))
        w = max(1.0, x1 - x0)
        short = (label[:24] + "…") if len(label) > 24 else label
        parts.append(
            f'<g class="bar-row">'
            f'<rect class="bar" x="{x0:.2f}" y="{y:.2f}" width="{w:.2f}" height="{bar_h:.2f}" rx="2" ry="2">'
            f'<title>{label}\n{c:,} rows ({p:.1f}%)</title>'
            f'</rect>'
            f'<text class="bar-label" x="{mleft-6}" y="{y + bar_h/2 + 3:.2f}" text-anchor="end">{short}</text>'
            f"<text class=\"bar-value\" x=\"{(x1 - 6 if w >= 56 else x1 + 4):.2f}\" y=\"{y + bar_h/2 + 3:.2f}\" text-anchor=\"{('end' if w >= 56 else 'start')}\">{c:,} ({p:.1f}%)</text>"
            f'</g>'
        )
    parts.append("</svg>")
    return "".join(parts)


def render_bool_card(s: Any) -> str:
    col_id = _safe_col_id(s.name)
    total = int(s.true_n + s.false_n + s.missing)
    cnt = int(s.true_n + s.false_n)
    miss_pct = (s.missing / max(1, total)) * 100.0
    miss_cls = 'crit' if miss_pct > 20 else ('warn' if miss_pct > 0 else '')
    true_pct_total = (s.true_n / max(1, total)) * 100.0
    false_pct_total = (s.false_n / max(1, total)) * 100.0

    mem_display = _human_bytes(getattr(s, 'mem_bytes', 0)) + ' (≈)'

    flags = []
    if miss_pct > 0:
        flags.append(f'<li class="flag {"bad" if miss_pct > 20 else "warn"}">Missing</li>')
    if cnt > 0 and (s.true_n == 0 or s.false_n == 0):
        flags.append('<li class="flag bad">Constant</li>')
    if cnt > 0:
        p = s.true_n / cnt
        if p <= 0.05 or p >= 0.95:
            flags.append('<li class="flag warn">Imbalanced</li>')
    quality_flags_html = f"<ul class='quality-flags'>{''.join(flags)}</ul>" if flags else ""

    left_tbl = f"""
    <table class=\"kv\"><tbody>
      <tr><th>Count</th><td class=\"num\">{cnt:,}</td></tr>
      <tr><th>Missing</th><td class=\"num {miss_cls}\">{s.missing:,} ({miss_pct:.1f}%)</td></tr>
      <tr><th>Unique</th><td class=\"num\">{(int(s.true_n>0)+int(s.false_n>0))}</td></tr>
      <tr><th>Processed bytes</th><td class=\"num\">{mem_display}</td></tr>
    </tbody></table>
    """
    right_tbl = f"""
    <table class=\"kv\"><tbody>
      <tr><th>True</th><td class=\"num\">{s.true_n:,} ({true_pct_total:.1f}%)</td></tr>
      <tr><th>False</th><td class=\"num\">{s.false_n:,} ({false_pct_total:.1f}%)</td></tr>
    </tbody></table>
    """

    def _bool_stack_svg(true_n: int, false_n: int, miss: int, width: int = 420, height: int = 48, margin: int = 4) -> str:
        total = max(1, int(true_n + false_n + miss))
        inner_w = width - 2 * margin
        seg_h = height - 2 * margin
        w_false = int(inner_w * (false_n / total))
        w_true = int(inner_w * (true_n / total))
        w_miss = max(0, inner_w - w_false - w_true)
        parts = [f'<svg class="bool-svg" width="{width}" height="{height}" viewBox="0 0 {width} {height}">']
        x = margin
        parts.append(f'<rect class="false" x="{x}" y="{margin}" width="{w_false}" height="{seg_h}"><title>False: {false_n:,}</title></rect>')
        x += w_false
        parts.append(f'<rect class="true" x="{x}" y="{margin}" width="{w_true}" height="{seg_h}"><title>True: {true_n:,}</title></rect>')
        x += w_true
        if w_miss:
            parts.append(f'<rect class="missing" x="{x}" y="{margin}" width="{w_miss}" height="{seg_h}"><title>Missing: {miss:,}</title></rect>')
        parts.append('</svg>')
        return ''.join(parts)

    chart_html = _bool_stack_svg(int(s.true_n), int(s.false_n), int(s.missing))
    safe_name = _html.escape(str(s.name))
    col_id = _safe_col_id(s.name)
    return f"""
    <article class=\"var-card\" id=\"{col_id}\">
      <header class=\"var-card__header\"><div class=\"title\"><span class=\"colname\">{safe_name}</span>
        <span class=\"badge\">Boolean</span>
        <span class=\"dtype chip\">{s.dtype_str}</span>
        {quality_flags_html}
      </div></header>
      <div class=\"var-card__body\">
        <div class=\"triple-row\">
          <div class=\"box stats-left\">{left_tbl}</div>
          <div class=\"box stats-right\">{right_tbl}</div>
          <div class=\"box chart\">{chart_html}</div>
        </div>
      </div>
    </article>
    """


def render_cat_card(s: Any) -> str:
    col_id = _safe_col_id(s.name)
    safe_name = _html.escape(str(s.name))
    total = s.count + s.missing
    miss_pct = (s.missing / max(1, total)) * 100.0
    miss_cls = 'crit' if miss_pct > 20 else ('warn' if miss_pct > 0 else '')
    approx_badge = '<span class="badge">approx</span>' if s.approx else ''

    mode_label, mode_n = (s.top_items[0] if s.top_items else ("—", 0))
    safe_mode_label = _html.escape(str(mode_label))
    mode_pct = (mode_n / max(1, s.count)) * 100.0 if s.count else 0.0

    import math
    if s.count > 0 and s.top_items:
        probs = [c / s.count for _, c in s.top_items]
        entropy = float(-sum(p * math.log2(max(p, 1e-12)) for p in probs))
    else:
        entropy = float('nan')

    rare_count = 0
    rare_cov = 0.0
    if s.count > 0:
        for _, c in s.top_items:
            pct = c / s.count * 100.0
            if pct < 1.0:
                rare_count += 1
                rare_cov += pct
    rare_cls = 'crit' if rare_cov > 60 else ('warn' if rare_cov >= 30 else '')
    top5_cov = 0.0
    if s.count > 0 and s.top_items:
        top5_cov = sum(c for _, c in s.top_items[:5]) / s.count * 100.0
    top5_cls = 'good' if top5_cov >= 80 else ('warn' if top5_cov <= 40 else '')
    empty_cls = 'warn' if s.empty_zero > 0 else ''

    flags = []
    if s.unique_est > max(200, int(0.5 * max(1, s.count))):
        flags.append('<li class="flag warn">High cardinality</li>')
    if mode_n >= int(0.7 * max(1, s.count)) and s.count:
        flags.append('<li class="flag warn">Dominant category</li>')
    if rare_cov >= 30.0:
        flags.append('<li class="flag warn">Many rare levels</li>')
    if s.case_variants_est > 0:
        flags.append('<li class="flag">Case variants</li>')
    if s.trim_variants_est > 0:
        flags.append('<li class="flag">Trim variants</li>')
    if s.empty_zero > 0:
        flags.append('<li class="flag">Empty strings</li>')
    if miss_pct > 0:
        flags.append(f'<li class="flag {"bad" if miss_pct > 20 else "warn"}">Missing</li>')
    quality_flags_html = f"<ul class=\"quality-flags\">{''.join(flags)}</ul>" if flags else ""

    mem_display = _human_bytes(int(getattr(s, 'mem_bytes', 0)))
    left_tbl = f"""
    <table class=\"kv\"><tbody>
      <tr><th>Count</th><td class=\"num\">{s.count:,}</td></tr>
      <tr><th>Unique</th><td class=\"num\">{s.unique_est:,}{' (≈)' if s.approx else ''}</td></tr>
      <tr><th>Missing</th><td class=\"num {miss_cls}\">{s.missing:,} ({miss_pct:.1f}%)</td></tr>
      <tr><th>Mode</th><td><code>{safe_mode_label}</code></td></tr>
      <tr><th>Mode %</th><td class=\"num\">{mode_pct:.1f}%</td></tr>
      <tr><th>Processed bytes</th><td class=\"num\">{mem_display} (≈)</td></tr>
    </tbody></table>
    """

    right_tbl = f"""
    <table class=\"kv\"><tbody>
      <tr><th>Entropy</th><td class=\"num\">{_fmt_num(entropy)}</td></tr>
      <tr><th>Rare levels</th><td class=\"num {rare_cls}\">{rare_count:,} ({rare_cov:.1f}%)</td></tr>
      <tr><th>Top 5 coverage</th><td class=\"num {top5_cls}\">{top5_cov:.1f}%</td></tr>
      <tr><th>Label length (avg)</th><td class=\"num\">{_fmt_num(s.avg_len)}</td></tr>
      <tr><th>Length p90</th><td class=\"num\">{s.len_p90 if s.len_p90 is not None else '—'}</td></tr>
      <tr><th>Empty strings</th><td class=\"num {empty_cls}\">{s.empty_zero:,}</td></tr>
    </tbody></table>
    """

    items = s.top_items or []
    maxN = max(1, min(15, len(items)))
    candidates = [5, 10, 15, maxN]
    topn_list = sorted({n for n in candidates if 1 <= n <= maxN})
    default_topn = 10 if 10 in topn_list else (max(topn_list) if topn_list else maxN)

    variants_html_parts = []
    for n in topn_list:
        if len(items) > n:
            keep = max(1, n - 1)
            head = items[:keep]
            other = sum(c for _, c in items[keep:])
            data = head + [("Other", other)]
        else:
            data = items[:n]
        svg = _build_cat_bar_svg_from_items(data, total=max(1, s.count + s.missing))
        active = " active" if n == default_topn else ""
        variants_html_parts.append(f'<div class="topn{active}" data-topn="{n}">{svg}</div>')
    topn_switch = " ".join(f'<button type="button" class="btn-soft btn-topn{(" active" if n == default_topn else "")}" data-topn="{n}">{n}</button>' for n in topn_list)
    chart_html = f"""
      <div class=\"topn-chart\">
        <div class=\"chart-variants\">{''.join(variants_html_parts)}</div>
        <div class=\"chart-controls\"><span>Top‑N:</span> {topn_switch}</div>
      </div>
    """

    return f"""
    <article class=\"var-card\" id=\"{col_id}\"> 
      <header class=\"var-card__header\"><div class=\"title\"><span class=\"colname\">{safe_name}</span>
        <span class=\"badge\">Categorical</span>
        <span class=\"dtype chip\">{s.dtype_str}</span>
        {approx_badge}
        {quality_flags_html}
      </div></header>
      <div class=\"var-card__body\">
        <div class=\"triple-row\">
          <div class=\"box stats-left\">{left_tbl}</div>
          <div class=\"box stats-right\">{right_tbl}</div>
          <div class=\"box chart\">{chart_html}</div>
        </div>
      </div>
    </article>
    """



from pysuricata.accumulators.boolean import BooleanAccumulator


def test_boolean_accumulator_counts():
    acc = BooleanAccumulator("flag")
    acc.update([True, False, None, True, False, False])
    s = acc.finalize()
    assert s.name == "flag"
    assert s.count == s.true_n + s.false_n
    assert s.missing >= 1
    # dtype string present
    assert s.dtype_str == "boolean"




"""Chunk consumption and accumulator wiring for polars chunks."""

from __future__ import annotations

from typing import Any, Dict, Iterable, List, Optional
import warnings

import numpy as np

try:  # optional
    import polars as pl  # type: ignore
except Exception:  # pragma: no cover
    pl = None  # type: ignore

from ..accumulators import (
    NumericAccumulator,
    BooleanAccumulator,
    DatetimeAccumulator,
    CategoricalAccumulator,
)
from .infer import ColumnKinds, infer_kind_for_series_polars


def _to_numeric_array_polars(s: "pl.Series") -> np.ndarray:  # type: ignore[name-defined]
    if pl is None:
        raise RuntimeError("polars not available")
    try:
        # Best-effort numeric casting; keep NaN for invalid
        s2 = s.cast(pl.Float64, strict=False)
        return s2.to_numpy()
    except Exception:
        # Fallback: to_list then numpy coercion
        return np.asarray(s.to_list(), dtype="float64")


def _to_bool_array_polars(s: "pl.Series") -> List[Optional[bool]]:  # type: ignore[name-defined]
    if pl is None:
        raise RuntimeError("polars not available")
    try:
        s2 = s.cast(pl.Boolean, strict=False)
        return [None if v is None else bool(v) for v in s2.to_list()]
    except Exception:
        out: List[Optional[bool]] = []
        for v in s.to_list():
            if v is None:
                out.append(None)
            else:
                vs = str(v).strip().lower()
                if vs in {"true", "1", "t", "yes", "y"}:
                    out.append(True)
                elif vs in {"false", "0", "f", "no", "n"}:
                    out.append(False)
                else:
                    out.append(None)
        return out


def _to_datetime_ns_array_polars(s: "pl.Series") -> List[Optional[int]]:  # type: ignore[name-defined]
    if pl is None:
        raise RuntimeError("polars not available")
    with warnings.catch_warnings():
        warnings.simplefilter("ignore", UserWarning)
        try:
            # Force ns resolution and get integer ns since epoch
            s2 = s.dt.cast_time_unit("ns") if s.dtype == pl.Datetime else s.cast(pl.Datetime("ns"), strict=False)
            # Cast to Int64 to extract raw ns (None for nulls)
            s3 = s2.cast(pl.Int64, strict=False)
            return [None if v is None else int(v) for v in s3.to_list()]
        except Exception:
            return [None] * len(s)


def _to_categorical_iter_polars(s: "pl.Series") -> Iterable[Any]:  # type: ignore[name-defined]
    return s.to_list()


def consume_chunk_polars(df: "pl.DataFrame", accs: Dict[str, Any], kinds: ColumnKinds, logger: Optional["logging.Logger"] = None) -> None:  # type: ignore[name-defined]
    if pl is None:
        raise RuntimeError("polars not available")

    # 1) Create accumulators for columns not seen in the first chunk
    for name in df.columns:
        if name in accs:
            continue
        kind = infer_kind_for_series_polars(df[name])
        if kind == "numeric":
            accs[name] = NumericAccumulator(name)
            kinds.numeric.append(name)
        elif kind == "boolean":
            accs[name] = BooleanAccumulator(name)
            kinds.boolean.append(name)
        elif kind == "datetime":
            accs[name] = DatetimeAccumulator(name)
            kinds.datetime.append(name)
        else:
            accs[name] = CategoricalAccumulator(name)
            kinds.categorical.append(name)
        if logger:
            logger.info("➕ discovered new column '%s' inferred as %s [pl]", name, kind)

    # 2) Feed accumulators for columns present in this chunk
    for name, acc in accs.items():
        if name not in df.columns:
            if logger:
                logger.debug("column '%s' not present in this chunk; skipping [pl]", name)
            continue
        s = df[name]
        if isinstance(acc, NumericAccumulator):
            arr = _to_numeric_array_polars(s)
            acc.update(arr)
            # extremes: approximate via argpartition like pandas path
            try:
                finite = np.isfinite(arr)
                if finite.any():
                    vals = arr[finite]
                    idx = np.arange(len(arr))[finite]
                    if vals.size > 0:
                        k = min(5, vals.size)
                        part_min = np.argpartition(vals, k - 1)[:k]
                        pairs_min = [(int(idx[i]), float(vals[i])) for i in part_min]
                        part_max = np.argpartition(-vals, k - 1)[:k]
                        pairs_max = [(int(idx[i]), float(vals[i])) for i in part_max]
                        acc.update_extremes(pairs_min, pairs_max)
            except Exception:
                pass
        elif isinstance(acc, BooleanAccumulator):
            acc.update(_to_bool_array_polars(s))
            try:
                acc.add_mem(int(df.select(pl.sum(pl.all().map_batches(lambda c: c.estimated_size()))).to_numpy()[0][0]))  # rough
            except Exception:
                pass
        elif isinstance(acc, DatetimeAccumulator):
            acc.update(_to_datetime_ns_array_polars(s))
            try:
                # rough memory accounting
                acc.add_mem(int(df.estimated_size()))
            except Exception:
                pass
        else:  # categorical
            acc.update(_to_categorical_iter_polars(s))




from __future__ import annotations

from dataclasses import dataclass, field
from typing import Any, Mapping, Optional, Sequence, Union, TYPE_CHECKING
import collections.abc as cabc
import os
import json

# Type-only imports so pandas/polars/pyarrow remain optional
if TYPE_CHECKING:  # pragma: no cover
    import pandas as pd  # type: ignore
    import polars as pl  # type: ignore

# Public data-like union: in-memory only (no file paths).
DataLike = Union[
    "pd.DataFrame",  # pandas
    "pl.DataFrame",  # polars eager
    "pl.LazyFrame",  # polars lazy
    cabc.Iterable,    # iterator/generator yielding pandas DataFrames
]


# Thin wrapper Report object with convenience methods
@dataclass
class Report:
    html: str
    stats: Mapping[str, Any]

    def save_html(self, path: str) -> None:
        with open(path, "w", encoding="utf-8") as f:
            f.write(self.html)

    def save_json(self, path: str) -> None:
        with open(path, "w", encoding="utf-8") as f:
            json.dump(self.stats, f, ensure_ascii=False, indent=2)

    def save(self, path: str) -> None:
        """Save based on extension (.html or .json)."""
        ext = os.path.splitext(path)[1].lower()
        if ext == ".html":
            self.save_html(path)
        elif ext == ".json":
            self.save_json(path)
        else:
            raise ValueError(f"Unknown extension for Report.save(): {ext}")

    # Jupyter-friendly inline display
    def _repr_html_(self) -> str:  # pragma: no cover - visual
        return self.html


@dataclass
class ComputeOptions:
    # General compute knobs
    chunk_size: Optional[int] = 200_000
    columns: Optional[Sequence[str]] = None
    numeric_sample_size: int = 20_000
    max_uniques: int = 2_048
    top_k: int = 50
    detect_pii: bool = False  # reserved for future integration
    engine: str = "auto"      # reserved
    # Parsing control (reserved)
    dtypes: Optional[Mapping[str, str]] = None
    random_seed: Optional[int] = 0


@dataclass
class RenderOptions:
    theme: str = "light"            # reserved; v2 uses built-in theme
    embed_assets: bool = True        # reserved; v2 HTML is self-contained
    show_quality_flags: bool = True  # reserved; chips already rendered


@dataclass
class ReportConfig:
    compute: ComputeOptions = field(default_factory=ComputeOptions)
    render: RenderOptions = field(default_factory=RenderOptions)


# Path-based loading is intentionally unsupported.


def _coerce_input(data: DataLike) -> Union["pd.DataFrame", cabc.Iterable]:
    """Resolve the actual input to feed into report_v2.

    Supports:
    - pandas.DataFrame
    - Iterator/generator yielding pandas DataFrames
    """
    # Path-like not supported
    # pandas DataFrame
    try:
        import pandas as pd  # type: ignore
        if isinstance(data, pd.DataFrame):
            return data  # type: ignore[return-value]
    except Exception:
        pass
    # Iterator/generator of DataFrames (duck-typed); let report_v2 validate on consumption
    try:
        if isinstance(data, cabc.Iterable) and not isinstance(data, (dict, list, tuple)):
            return data  # type: ignore[return-value]
    except Exception:
        pass
    raise TypeError(
        "Unsupported data type for this API. Provide a pandas DataFrame, a polars DataFrame, or an iterable of pandas DataFrames."
    )


def _to_engine_config(path_or_df: Union["pd.DataFrame", cabc.Iterable], cfg: ReportConfig) -> "report_v2.ReportConfig":  # type: ignore[name-defined]
    """Translate our high-level ReportConfig into engine config (internal)."""
    from . import report_v2

    compute = cfg.compute
    # Base mapping
    v2 = report_v2.ReportConfig(
        chunk_size=compute.chunk_size or 200_000,
        numeric_sample_k=compute.numeric_sample_size,
        uniques_k=compute.max_uniques,
        topk_k=compute.top_k,
        engine=compute.engine,
    )

    return v2


def profile(
    data: DataLike,
    config: Optional[ReportConfig] = None,
) -> Report:
    """Compute stats and render HTML. Returns a Report(html, stats).

    - data: In-memory DataFrame or iterable of DataFrames.
    - config: High-level config with compute/render options.
    - loader: deprecated (not supported)
    """
    from . import report_v2

    cfg = config or ReportConfig()
    inp_raw = _coerce_input(data)
    # Special-case: in-memory polars -> chunked pandas frames
    wrapped = None
    try:
        import polars as pl  # type: ignore
        if isinstance(inp_raw, (pl.DataFrame,)) or getattr(inp_raw, "__class__", None).__name__ == "LazyFrame":
            from .io import iter_chunks as _iter_chunks
            wrapped = _iter_chunks(
                inp_raw, chunk_size=cfg.compute.chunk_size, columns=cfg.compute.columns
            )
    except Exception:
        pass
    inp = wrapped if wrapped is not None else inp_raw
    v2cfg = _to_engine_config(inp, cfg)

    # Always compute stats to return machine-readable mapping
    html, summary = report_v2.build_report(inp, config=v2cfg, return_summary=True)  # type: ignore[misc]

    try:
        stats = dict(summary or {})
    except Exception:
        stats = {"dataset": {}, "columns": {}}
    return Report(html=html, stats=stats)


def summarize(
    data: DataLike,
    config: Optional[ReportConfig] = None,
) -> Mapping[str, Any]:
    """Stats-only manifest. JSON-safe mapping suitable for CI/data-quality checks."""
    from . import report_v2
    cfg = config or ReportConfig()
    inp_raw = _coerce_input(data)
    # Special-case: in-memory polars -> chunked pandas frames
    wrapped = None
    try:
        import polars as pl  # type: ignore
        if isinstance(inp_raw, (pl.DataFrame,)) or getattr(inp_raw, "__class__", None).__name__ == "LazyFrame":
            from .io import iter_chunks as _iter_chunks
            wrapped = _iter_chunks(
                inp_raw, chunk_size=cfg.compute.chunk_size, columns=cfg.compute.columns
            )
    except Exception:
        pass
    inp = wrapped if wrapped is not None else inp_raw
    v2cfg = _to_engine_config(inp, cfg)
    # compute-only to skip HTML render
    _html, summary = report_v2.build_report(inp, config=v2cfg, return_summary=True, compute_only=True)  # type: ignore[misc]
    stats = dict(summary or {})
    return stats
